 /***************************************************************************
                          ftpfs.h  -  description
                             -------------------
    begin                : sat nov 30 2002
    copyright            : (C) 2002 by Piotr Mierzwiski
    email                : peterm@go2.pl

    copyright            : See COPYING file that comes with this project

$Id: ftpfs.h 641 2009-06-30 20:38:49Z piotr $
 ***************************************************************************/

#ifndef _FTPFS_H_
#define _FTPFS_H_

#include "enums.h"
#include "virtualfs.h"

#include <qftp.h>
#include <qptrstack.h>

/**
 * Author:
 * Piotr Mierzwiski
 **/
/** @short Klasa klienta FTP.
 Obsugiwane s tu wszystkie podstawowe operacje na plikach, takie jak:
 kopiowanie/przenoszenie usuwanie, tworzenie nowych pustych, tworzenie nowych
 katalogw, zmiana atrybutw. Podczas listowania rozrniane s linki do plikw
 i katalogw. Nie s wspierane rekurencyjne operacje na plikach.
 */
class FtpFS : public VirtualFS
{
	Q_OBJECT
public:
	/** Konstruktor klasy.\n
	 @param inputURL - nazwa biecej cieki,
	 @param listViewExt - wskanik na obiekt widoku listy plikw.\n
	 Inicjowane s tu skadowe klasy, tworzony obiekt klasy QFtp oraz czone
	 sygnay ze slotami.
	 */
	FtpFS( const QString & inputURL, ListViewExt *listViewExt );

	/** Destruktor.\n
	 Usuwany jest tu obiekt uzywany w klasie (klasy QFtp).
	 */
	~FtpFS();

	/** Zwraca nazw gwnego katalogu.\n
	 @return nazwa gwnego katalogu - 'ftp://nazwa_hosta/'.
	 */
	QString host() const { return QString("ftp://"+mHostName+"/"); }

	/** Zwraca biec ciek absolutn.\n
	 @return bieca cieka absolutna (bez nazwy gwnego katalogu).
	 */
	QString path() const { return mRemoteDir; }

	/** Zwraca numer portu na ktrym obsugiwany jest klient FTP.\n
	 @return numer portu.
	 */
	uint portNum() const { return mPortNum; }

	/** Zwraca nazw (ze ciek absolut) aktualnie przetwarzanego pliku.
	 @return nazwa pliku ze cie absolutn.
	 */
	QString nameOfProcessedFile() const { return mNameOfProcessedFile; }

	/** Zwraca kod bdu dla ostatniej operacji plikowej.
	 @return kod bdu, patrz: @see Error.
	 */
	Error errorCode() const { return mErrorCode; }

	/** Uruchamia funkcj pobierajc wyniki operacji waenia, tj. waga caoci,
	 ilo plikw i katalogw.\n
	 @param weight - zmienna, w ktrej zapisana bdzie waga,
	 @param files - zmienna do zapisania iloci plikw,
	 @param dirs - zmienna do zapisania iloci katalogw.\n
	 Uwaga! Metoda nie ma definicji.
	 */
	void getWeighingStat( long long & weigh, uint & files, uint & dirs );

	/** Pobiera informacj o pliku wczeniej zainicjowan w skadowej
	 @em mNewUrlInfo .\n
	 @param newUrlInfo - zmienna, w ktrej naley umieci informacj o pliku.\n
	 Metoda jest wykorzystywana przez @em FilesPanel::slotResultOperation() ,
	 w celu zapewnienia odpowiedniej obsugi widoku listy.
	 */
	void getNewUrlInfo( UrlInfoExt & newUrlInfo ) { newUrlInfo = mNewUrlInfo; }

	/** Metoda zwraca pocztkow pozycj nazwy w listowanym elemencie.
	 @return indeks nazwy.
	 */
	int nameIndex( const QString & name );

	/** Funkcja zwraca status zachowania fragmentu kopiowanego lub przenoszonego pliku.
	@return TRUE jezeli zachowano czesc pliku, w przeciwnym razie FALSE.
	 */
	bool afterBreakKeepFragment() { return m_bAfterBreakKeepFragment; }

private:
	QFtp *mFtp;

	QString mHostName, mRemoteDir, mUserName, mPassword;
	QString mOldUserName, mOldPassword;
	QFtp::Command mLastFinishedCmd;
	Error mErrorCode;
	uint mPortNum, mOldPortNum;

	QString mNameOfProcessedFile, mTargetPath;
	QStringList mCopiedFilesList;

	uint mOpFileCounter, mTotalFiles;
	bool mMoveFiles;

	bool mPassiveModeOn;
	bool mUploading;
	bool mGettingURL;
	UrlInfoExt mNewUrlInfo;
	int mStat;

	bool mReadFile, mFileIsReadable, mSkipNextErrorOccures;
	QPtrList<QFile> mOpenedFilesPtrList;

	QStringList mMainFilesList, mTargetFileForLinkLst;
	uint mLinkCheckingCounter;
	bool mLinkChecking;

	bool m_bAfterBreakKeepFragment;

	uint mRealWritedBytes;

	QTimer *m_pTimer;

protected:
	//bool getFtpAttributs( const QString & inputURL );

	/** Metoda uruchamia procedur poczenia z podanym URLem.\n
	 @param inputURL - URL na ktry naley si poczy,
	 @param error - kod bdu (wczeniejszej operacji).
	 */
	void connectToHost( const QString & inputURL, QFtp::Error error=QFtp::NoError );

	/** Powoduje zmian biecego katalogu na podany.\n
	 @param dirName - nowa (wzgldna) cieka do katalogu.
	 */
	void changeDirOn( const QString & dirName );


	/** Zamyka biece poczenie jeli ono istnieje, lub jest wymuszane.\n
	 @param force TRUE, wymusza wykonanie rozczenia z hostem, jeli FALSE
	 wtedy nastpi rozczenie tylko wtedy jeli poczenie takie istnieje.
	 */
	void closeCurrentConnection( bool force=FALSE );

	/** Funkcja uruchamia procedur przerwania biecej operacji.\n
	 UWAGA! Narazie wywoywany jest tu @see slotDone oraz ukrywany jest dialog
	 postpu wykonywania operacji.
	 */
	void breakCurrentOperation();


	/** Sprawdza obecno pliku na serwerze wykonuj operacj STAT na jego rzecz.\n
	 @param fileName - nazwa pliku do otwarcia.\n
	 Uwaga: metod naley zawsze wywoywa przed otwarciem pliku (do podgldu).
	 */
	void open( const QString & fileName );


	/** Jeli uytkownik jest zalogowany, wtedy podejmowana jest prba zmiany
	 katalogu, w przeciwnym razie pokazywane jest okienko z pytaniem o ponowne
	 poczenie.\n
	 @param dirName - nowy katalog do odczytania.\n
	 Odpowied 'yes' w oknie dialogowym poczenia uruchomi poczenie, a jeli
	 ono zostanie zakoczone sukcesem wtedy nastpi otwarcie wybranego katalogu.
	*/
	void readDir( const QString & dirName );

	/** Utworzenie nowego katalogu.\n
	 @param dirName - nazwa nowego katalogu.
	 */
	void mkdir( const QString & dirName );

	/** Metoda uruchamia procedur utworzenia nowego pliku lub katalogu.\n
	 @param fullFileName - nazwa pliku lub katalogu do utworzenia.\n
	 Jeli nazwa koczy sie znakiem '/', wtedy wywoywane jest polecenie
	 tworzenia nowego katalogu, w przeciwnym razie tworzony jest pusty plik.
	*/
	void createAnEmptyFile( const QString & fileName );

	/** Uruchamiane jest tutaj kopiowanie lub przenoszenie podanych plikw do
	 podanego katalogu (z FTP na LOCALfs i odwrotnie).\n
	 @param targetPath - miejsce docelowe kopiowania/przenoszenia,
	 @param selectedFilesList - lista plikw do skopiowania/przeniesienia,
	 @param savePerm - TRUE, wymusza zachowanie atrybutw kopiowanych plikw
	 @param alwaysOvr - TRUE mwi, e pliki docelowe bd nadpisywane
	 @param followByLink - TRUE, wymusza kopiowanie dowiza linkw, FALSE
	  spowoduje kopiowanie linkw jako takich.
	 @param move - TRUE, oznacza e pliki bd przenoszone, w przeciwnym razie
	  zostan skopiowane.\n
	 Przy czym obsugiwane jest tylko kopiowanie plikw (katalogi kopiowane s
	 tylko jako puste).
	 */
	void copyFiles( QString & targetPath, const QStringList & selectedFilesList, bool savePerm, bool alwaysOvr, bool followByLink, bool move );

	/** Metoda powoduje podjcie prby usunicia podanych plikw.\n
	 @param filesList - lista plikw do usunicia.
	 */
	void deleteFiles( const QStringList & filesList );


	/** Metoda uruchamia "surowe polecenie" ustawiania atrybutw dla kadego pliku
	 z listy.\n
	 @param filesList - lista plikw, ktrym naley ustawi atrybuty,
	 @param newUrlInfo - informacja o atrybutach dla plikw,
	 @param recursive - rwne TRUE, spowoduje e atrybuty bd zmieniane dla caej
	  zawartoci katalogw z listy,
	 @param changesFor - rodzaj zmiany atrybutw, dozwolone s:
	  @li 0 - zmiana dla wszystkich,
	  @li 1 - zmiana tylko dla plikw,
	  @li 2 - zmiana tylko dla katalogw.\n
	 Wynik dziaania polecenia naley sprawdza w @see slotRawCommandReply.\n
	 UWAGA! Zmiana atrybutw jest narazie wykonywana tylko dla pierwszego pliku
	 z listy.
	 */
	void setAttributs( const QStringList & filesList, const UrlInfoExt & newUrlInfo, bool recursive, int changesFor );
//	void rename( const QString & oldName, const QString & newName ) { mFtp->rename( oldName, newName ); }

	/** Metoda uruchamia waenie podanych na licie plikw.\n
	 @param filesList - lista plikw do zwaenia
	 @param realWeight - rwne TRUE spowoduje, e funkcja waca uwzgldni
	  rzeczywist wag pliku w systemie plikw.\n
	 UWAGA! Brak definicji.
	 */
	void sizeFiles( const QStringList & filesList, bool realWeigh );

	/** Funkcja wyuskuje z podanego cigu poszczaglne skadowe adresu.\n
	 @param path - nowa cieka.\n
	 Nastpuje tu wycignicie s nastpujce skadowe adresu: nazwa hosta,
	 zdalny katalog, numer portu, uytkownik oraz haso, po czym zainicjowanie
	 skadowych klasy.\n \n Przykady moliwych cieek wejciowych:\n
	  @li "ftp://user:password@hostname.com:21/subdir/"
	  @li "ftp://hostname.com/subdir/"
	*/
	void initSessionParameters( const QString & path );

	/** Zwraca informacj o tym czy uytkownik jest aktualnie zalogowany.\n
	 @return TRUE oznacza, e uytkownik jest zalogowany, FALSE - nie jest
	 zalogowany.
	 */
	bool isLoggedIn() { return (mFtp->state() == QFtp::LoggedIn); }


	/** Metoda wykonuje translacj metody typu QFtpCommand na Operation.
	 @return polecenie typu @em Operation .\n
	 Przy czym polecenie QFtpCommand zawiera skadowa @em mLastFinishedCmd .
	 */
	Operation currentOperation() const;

	/** Powoduje pokazanie okna dialogowego z pytaniem o ponowne poczenie
	 z hostem.\n
	 @return TRUE w przypadku udanego poczenia, w przeciwnym razie FALSE.\n
	 Jeli wybrane zostanie 'Yes', wtedy nastpuje uruchomienie poczenia.
	 */
	bool connectAgain();


	/** Zwraca nazw uytkownika, ktry jest aktualnie zalogowany.\n
	 @return nazwa uytkownika
	 */
	QString loggedOwner() { return mUserName; }


	/** Uruchamia polecenie usunicia podanego pliku z lokalnego systemu plikw.\n
	 @param fileName - nazwa pliku do usunicia.
	 @return TRUE, jeli operacja si powiedzie, w przeciwnym razie FALSE.\n
	 Jeli pliku nie uda si usun pokazywane jest okno informacyjne.
	 */
	bool removeFileFromLocalFS( const QString & fileName );

	/** Metoda uruchamia "surowe polecenie" sprawdzenia, kadego pliku docelowego
	 dla linkw z listy zebranej podczas listowania plikw.
	 ( @em mTargetFileForLinkLst ).\n
	 Wynik dziaania polecenia naley sprawdza w @see slotRawCommandReply.
	 */
	void checkAllLinks();

public slots:
	/** Slot powoduje zainicjowanie podanego obiektu z informacj o pliku
	 atrybutami podanego.\n
	 @param fullFileName - nazwa pliku, ktremu naley pobra atrybuty,
	 @param urlInfoExt - adres obiektu, ktry naley zainicjowa atrybutami.\n
	 Przy czym pobierane s tylko nastpujce atrybuty: nazwa, rozmiar oraz czas
	 ostatniej modyfikacji podanego pliku.
	 */
	void slotGetUrlInfo( const QString & fullFileName, UrlInfoExt & urlInfoExt );

	/** Slot powoduje wczytywanie pliku wybranego do podgldu.\n
	 @param fileName - nazwa pliku do wczytania,
	 @param buffer - adres bufora, w ktrym naley umieszcza wczytywane dane,
	 @param bytesRead - ilo wczytywanych na raz bajtw, a po odczycie zawiera
	  ilo faktycznie wczytanych bajtw (0 oznacza koniec pliku)
	 @param fileLoadMethod - metoda wczytywania pliku, patrz @see View::UpdateMode
	 */
	void slotReadFileToView( const QString & fileName, QByteArray & buffer, int & bytesRead, int fileLoadMethod );

private slots:
	/** Pokazywane sa tutaj bledy dla ostatnio wykonywanej czynnosci.
	 @param finishedWithError - TRUE, oznacza e operacja zakoczya si bdem
	  (do odczytania przez @em errorCode() ), FALSE mowi e wszystko poszo OK.
	*/
	void slotDone( bool finishedWithError );

	/** Slot powoduje wywietlenie informacje o stanie biecej operacji na pasku
	 statusu.\n
	 @param state - status operacji (typu QFtp::State).\n
	 Wysyany jest tutaj tylko sygna @em signalSetOperation() .
	 */
	void slotStateChanged( int state );

	/** Slot wywoywany w momencie rozpoczcia wykonywania polecenia.\n
	 @param cmd - kod polecenia (typu QFtp::Command).\n
	 Wykonywane s tu czynnoci przygotowujce, niezbdne dla danego polecenia.
	 */
	void slotCommandStarted( int command );

	/** Slot wywoywany w chwili zakoczenia wykonywania podanego polecenia.\n
	 @param cmd - kod polecenia ( typu @em Operation ),
	 @param error - rwne TRUE oznacza, e pojawi si bd w wykonanym poleceniu.\n
	 Jeli prba wykonania polecenia zakoczya si bdem pokazywany jest
	 komunikat z informacj o bdzie i nazwy pliku lub katalogu, ktrego bd
	 dotyczny.
	 */
	void slotCommandFinished( int command, bool error );

	/** W slocie parsowane s polecenia uruchaniane "rcznie" na hocie FTP.\n
	 @param code - kod polecenia
	 @param text - wynik wykonania polecenia
	 */
	void slotRawCommandReply( int code, const QString & text );

	/**
	W slocie wysyane s sygnay z informacjami postpie danej operacji.\n
	S to:\n
	 @see signalFileCounterProgress()\n
	 @see signalTotalProgress()\n
	 @see signalDataTransferProgress().
	*/
	void slotSetProgress( int done, int total );


	/** Slot wywoywany przez obiekt klasy FileView, po wykonaniu zapisu do pliku.\n
	 @param fileName - nazwa zapisywanego pliku.
	 */
	void slotUpdateItemOnLV( const QString & fileName );

	void slotQFtpListInfo( const QUrlInfo & ui );

	/** Slot wywoywany po upyniciu okrelonego czasu.
	 * Uywany do dziaania trybu podtrzymywania poczenia.
	*/
	void slotTimeOut();

signals:
	/** Sygna wymusza pokazanie dialogu nadpisywania pliku.\n
	 @param sourceFileName - nazwa rdowego pliku,
	 @param targetFileName - nazwa docelowego pliku,
	 @param result - zmienna zawiera kod odpowiedzi uytkownika.
	 */
	void signalFileOverwrite( const QString & sourceFileName, QString & targetFileName, int & result );

	/** Sygna zawiera informacj o aktualnie listowanym elemencie.
	 @param urlInfoExt - informacja o pliku.\n
	 Wysyany jest podczas listowania katalogu.
	 */
	void listInfo( const UrlInfoExt & urlInfoExt );

};

#endif
