/***************************************************************************
                          keyshortcuts.h  -  description
                             -------------------
    begin                : sun aug 1 2004
    copyright            : (C) 2004 by Piotr Mierzwiski
    email                : peterm@go2.pl

    copyright            : See COPYING file that comes with this project
 ***************************************************************************/

#ifndef _KEYSHORTCUT_H_
#define _KEYSHORTCUT_H_

#include <qobject.h>
#include <qvaluelist.h>

/** @short Klasa pomocnicza trzymajca opis pojedynczego kodu klawiszowego.
 Na opis skrtu skadaj si nastpujce rzeczy: identyfikator akcji, opis
 akcji, domylny kod klawisza, aktualny kod klawisza. W celu utworzenia kodu
 naley zainicjowa konstruktor. Klasa udostpnia niezbdne metody do
 pobierania i ustawiania wartoci kodu, przy czym ustawi mona tylko kod
 klawiszowy.
 */
class KeyCode
{
public:
	/** Konstruktor.\n
	 @param act - identyfikator akcji,
	 @param desc - opis akcji,
	 @param dfltCode - domylny kod klawisza,
	 @param _code - waciwy kod klawisza.\n
	 Inicjalizowane s tutaj skadowe klasy.
	 */
	KeyCode( int act, const QString & desc, int dfltCode, int _code=Qt::Key_unknown ) :
		mAction(act), mDescription(desc), mDefaultCode(dfltCode), code(_code)
	{}

	/** Destruktor.\n
	 Nie zawiera definicji.
	 */
	~KeyCode() {}

	/** Zwraca opis akcji.\n
	 @return opis akcji.
	 */
	QString description() const { return mDescription; }

	/** Zwraca domylny kod klawisza.\n
	 @return domylny kod klawisza.
	 */
	int defaultCode() const { return mDefaultCode; }

	/** Zwraca identyfikator akcji.\n
	 @return identyfikator akcji.
	 */
	int action() const { return mAction; }

private:
	int mAction;
	QString mDescription;
	int mDefaultCode;
	int mCode;

public:
	/** Aktualny kod klawisza.
	 */
	int code;

};

// ooooooooooo-----------         -------------ooooooooooo

/** @short Klasa obsugujca skrty klawiszowe.
 Klasa przeznaczona do edycji skrtw klawiszowych uywanych w oknie podgldu
 oraz oknie gwnym aplikacji. Trzymana jest tutaj lista powyszych,
 udostpniane s metody konieczne do pobierania i ustawiania skrtw. W
 momencie inicjacji nastpuje prba odczytania ich z pliku konfiguracyjnego,
 jeli si to nie uda przyjmowane s domylne wartoci.
*/
class KeyShortcuts : public QObject
{
	Q_OBJECT
public:
	/** Konstruktor.\n
	 Czyszczona jest tutaj tylko lista skrtw.
	 */
	KeyShortcuts();

	/** Destruktor.\n
	 Nie zawiera definicji.
	 */
	virtual ~KeyShortcuts() {}

	/** Powoduje uaktualnienie kodw skrtw klawiszowych na podstawie klucza z
	 pliku konfiguracyjnego.\n
	 @param configFileKey - klucz dla pliku konfiguracyjnego.\n
	 Przy czym przed wywoaniem tej funkcji naley wypeni list domylnymi
	 wartociami, patrz: @see append.
	 */
	virtual void updateEntryList( const QString & configFileKey );

	/** Zwraca skrt klawiszowy dla podanego jego identyfikatora akcji.\n
	 @param action - identyfikator akcji dla skrtu klawiszowego, (patrz te:
	  @see ActionsOfView, oraz @see ActionsOfApp),
	 @return skrt klawiszowy.
	 */
	virtual int key( int action );

	/** Metoda powoduje ustawienie podanego skrtu klawiszowego dla podanego
	 jego opisu.\n
	 @param action - identyfikator akcji dla skrtu klawiszowego, (patrz te:
	  @see ActionsOfView, oraz @see ActionsOfApp),
	 @param shortcut - nowy skrt do ustawienia.
	 */
	virtual void setKey( int action, int shortcut );

	/** Metoda powoduje ustawienie podanego skrtu klawiszowego dla podanego
	 jego identyfikatora akcji.\n
	 @param description - opis akcji dla skrtu klawiszowego,
	 @param shortcut - nowy skrt do ustawienia.
	 */
	virtual void setKey( const QString & description, int shortcut );


	/** Zwraca aktualn ilo skrtw klawiszowych.\n
	 @return liczba skrtw klawiszowych.
	 */
	uint count() const { return mKeyCodeList.count(); }

	/** Metoda powoduje wyczyszczenie listy skrtw klawiszowych.
	 */
	void clear() { mKeyCodeList.clear(); }

	/** Metoda umoliwia dodanie skrtu klawiszowego do listy.\n
	 @param action - identyfikator akcji,
	 @param description - opis akcji,
	 @param defaultKeyShortcut - domylny skrt klawiszowy. Przy czym w przypadku
	 kombinacji klawiszowych naley uy typu 'Qt::Modifier', np. Qt::CTRL.
	 */
	virtual void append( int action, const QString & description, int defaultKeyShortcut=Qt::Key_unknown ) {
		mKeyCodeList.append( new KeyCode(action, description, defaultKeyShortcut) );
	}

	/** Wykonuje konwersje z typu QKeyEvent to QKeySequence i zwraca ten drugi.\n
	 @param ke - wskanika na zdarzenie KeyEvent,
	 @return zwraca skrt klawiszowy typu QKeySequence.
	 */
	virtual int ke2ks( QKeyEvent *ke );

	/** Zwraca cig tekstowy domylnego skrtu klawiszowego dla podanego Id.\n
	 @return cig tekstowy opisujcy skrt klawiszowy.
	 */
	virtual QString keyDefaultStr( uint index ) const;

	/** Zwraca cig tekstowy aktualnego skrtu klawiszowego dla podanego Id.\n
	 @return cig tekstowy opisujcy skrt klawiszowy.
	 */
	virtual QString keyStr( uint index ) const;

	/** Zwraca opis akcji dla podanego Id.
	 @return opis akcji.
	 */
	QString desc( uint index ) const {
		return (*mKeyCodeList.at( index ))->description();
	}

private:
	typedef QValueList <KeyCode *> KeyCodeList;
	KeyCodeList mKeyCodeList;

};

#endif
